@@ -54,6 +54,10 @@
  */
 #define PROCESS_OUTSTANDING_TIMEOUT 7200
 
+/* The maximum allowed size of a report file, if not overridden
+ */
+#define MAX_REPORT_FILESIZE 1000000000
+
 /* If true, we have an active Internet connection. True by default in case we
  * can't bring up GNetworkMonitor */
 static gboolean online_state = TRUE;
@@ -198,6 +202,9 @@ static const char* unacceptable_fields[] = {
 
     "Title",
 
+    /* Redundant with the addition of pa-info LP: #1893899 */
+    "AlsaInfo",
+
     NULL
 };
 
@@ -261,7 +268,7 @@ split_string (char* head, char** tail)
 
 gboolean
 bsonify (GHashTable* report, bson* b, const char** bson_message,
-         int* bson_message_len)
+         uint32_t* bson_message_len)
 {
     /* Attempt to convert a #GHashTable of the report into a BSON string.
      * On error return %FALSE. */
@@ -295,7 +302,7 @@ bsonify (GHashTable* report, bson* b, const char** bson_message,
 }
 
 int
-upload_report (const char* message_data, int message_len, GString *s)
+upload_report (const char* message_data, uint32_t message_len, GString *s)
 {
     CURL* curl = NULL;
     CURLcode result_code = 0;
@@ -343,14 +350,21 @@ upload_report (const char* message_data, int message_len, GString *s)
         return response_code;
 }
 
-void
-destroy_key_and_value (gpointer key, gpointer value, gpointer user_data)
+gsize
+get_report_max_size (void)
 {
-    if (key)
-        g_free (key);
-    /* The value may be "", which is allocated on the stack. */
-    if (value && *(char*)value != '\0')
-        g_free (value);
+    const char* value = NULL;
+    gsize max_size;
+
+    value = g_getenv ("REPORT_MAX_SIZE");
+    if (value == NULL)
+        return MAX_REPORT_FILESIZE;
+
+    max_size = atol(value);
+    if (max_size == 0)
+        return MAX_REPORT_FILESIZE;
+
+    return max_size;
 }
 
 GHashTable*
@@ -371,22 +385,48 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
     gchar* token_p = NULL;
     char* key = NULL;
     char* value = NULL;
+    char* old_value = NULL;
     gchar* value_p = NULL;
     GError* err = NULL;
     gchar* end = NULL;
     size_t value_length;
     size_t value_pos;
+    struct stat st;
+    int fd;
+    int res;
+    gsize max_size;
 
     g_return_val_if_fail (report_path, NULL);
 
-    if (g_file_test (report_path, G_FILE_TEST_IS_SYMLINK) ||
-        !g_file_test (report_path, G_FILE_TEST_IS_REGULAR)) {
+    max_size = get_report_max_size();
+
+    fd = open(report_path, O_RDONLY | O_NOFOLLOW);
+    res = fstat(fd, &st);
+    if (res == -1) {
         g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
-                     "%s is a symlink or is not a regular file.", report_path);
+                     "%s could not be opened.", report_path);
+        if (fd >= 0)
+            close(fd);
         return NULL;
     }
+
+    if(!S_ISREG(st.st_mode)) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is not a regular file.", report_path);
+        close(fd);
+        return NULL;
+    }
+
+    /* Limit the size of a report */
+    if(st.st_size > max_size) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is too big to be processed.", report_path);
+        close(fd);
+        return NULL;
+    }
+
     /* TODO handle the file being modified underneath us. */
-    fp = g_mapped_file_new (report_path, FALSE, &err);
+    fp = g_mapped_file_new_from_fd (fd, FALSE, &err);
     if (err) {
         g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                      "Unable to map report: %s", err->message);
@@ -396,8 +436,17 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
 
     contents = g_mapped_file_get_contents (fp);
     file_len = g_mapped_file_get_length (fp);
+
+    /* Check report size again to make sure */
+    if(file_len > max_size) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is too big to be processed.", report_path);
+        goto error;
+    }
+
     end = contents + file_len;
-    hash_table = g_hash_table_new (g_str_hash, g_str_equal);
+    hash_table = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                        g_free, g_free);
     p = contents;
 
     while (p < end) {
@@ -436,9 +485,15 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                 value_p = value + value_pos;
                 *value_p = '\n';
                 value_p++;
+                g_hash_table_steal (hash_table, key);
             } else {
                 value = g_realloc (value, value_length + 1);
                 value_p = value;
+                /* Make sure we properly free the old empty string value */
+                old_value = g_hash_table_lookup (hash_table, key);
+                if (old_value)
+                    g_free (old_value);
+                g_hash_table_steal (hash_table, key);
             }
             memcpy (value_p, p, value_length);
             value_p[value_length] = '\0';
@@ -447,7 +502,7 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                 if (*c != '\t' && *c >= '\0' && *c < ' ')
                     *c = '?';
             value_p += value_length;
-            g_hash_table_insert (hash_table, key, value ? value : "");
+            g_hash_table_insert (hash_table, key, value ? value : g_strdup(""));
             p = token_p + 1;
         } else {
             /* Reset the value pointer. */
@@ -517,10 +572,11 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
             }
             p = token_p + 1;
 
-            g_hash_table_insert (hash_table, key, value ? value : "");
+            g_hash_table_insert (hash_table, key, value ? value : g_strdup(""));
         }
     }
     g_mapped_file_unref (fp);
+    close(fd);
 
     /* Remove entries that we don't want to send */
     if (!full_report) {
@@ -538,8 +594,7 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                     continue;
             }
 
-            g_hash_table_iter_steal(&iter);
-            destroy_key_and_value(key, value, NULL);
+            g_hash_table_iter_remove(&iter);
         }
     }
 
@@ -547,10 +602,10 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
 
 error:
     if (hash_table) {
-        g_hash_table_foreach (hash_table, destroy_key_and_value, NULL);
         g_hash_table_destroy (hash_table);
     }
     g_mapped_file_unref (fp);
+    close(fd);
     return NULL;
 }
 
@@ -665,9 +720,8 @@ parse_and_upload_report (const char* crash_file)
 {
     GHashTable* report = NULL;
     gboolean success = FALSE;
-    int message_len = 0;
+    uint32_t message_len = 0;
     const char* message_data = NULL;
-    GString* s = NULL;
     GError* error = NULL;
     bson b[1];
     int response = 0;
@@ -694,7 +748,7 @@ parse_and_upload_report (const char* crash_file)
         success = TRUE;
     } else {
         log_msg ("Uploading %s.\n", crash_file);
-        s = g_string_new (NULL);
+        GString* s = g_string_new (NULL);
         response = upload_report (message_data, message_len, s);
         if (bson_data (b))
             bson_destroy (b);
@@ -715,10 +769,10 @@ parse_and_upload_report (const char* crash_file)
 
         if (response == 200 && s->len > 0)
             handle_response (report, s->str);
+
+        g_string_free (s, TRUE);
     }
 
-    g_string_free (s, TRUE);
-    g_hash_table_foreach (report, destroy_key_and_value, NULL);
     g_hash_table_destroy (report);
 
     return success;
